/// File: P5GloveSharp\P5Glove\P5Glove.h
/// 
/// ------------------------------------------------------------
/// Copyright (c) 2004
///   Antonio Cisternino (cisterni.unipi.it),
///   Diego Colombo      (colombod.unipi.it),
///   Giorgio Ennas      (   ennas.unipi.it),
///   Daniele Picciaia   (picciaia.unipi.it)
/// 
/// The use and distribution terms for this software are 
/// contained in the file named license.txt, which can be found 
/// in the root of this distribution.
/// By using this software in any fashion, you are agreeing to 
/// be bound by the terms of this license.
///
/// You must not remove this notice, or any other, from this
/// software.
/// ------------------------------------------------------------
/**
 * $Log: /Common/P5GloveSharp/P5Glove/P5Glove.h $
 * 
 * 5     5/13/05 3:39p Eaddy
 * o Fixed warnings
 * 
 * 4     5/13/05 3:18p Eaddy
 * o Fixed bug where IsGesture always returned false
 */

// P5Glove.h

#pragma once

using namespace System;
using namespace System::Drawing;
using namespace System::Windows::Forms;

namespace P5
{
	// Bit flags
	public __value enum Fingers
	{
		// l = straight
		// r = partly bent
		// n = completely bent
		// x = don't care
		//
		// You use a sequence of five characters starting with the thumb.
		//
		// eg.
		//
		// pointing = xlnnn
		// fist = nnnnn
		// claws = rrrrr
		//
		// Acknowledgements: Carl Kenner came up with the clever lrnx
		// designation.

		DontCare	= 0x00000000,	// x
		Straight	= 0x00000001,	// l
		PartlyBent  = 0x00000002,	// r
		TotallyBent = 0x00000003,	// n

		Thumb		= 0x00010000,
		Index		= 0x00020000,
		Middle		= 0x00040000,
		Ring		= 0x00080000,
		Pinky		= 0x00100000,
	};

	// Example: Fist: nnnnn, SplayedHand: lllll, LeBird: rrlrr
	static const int nDONT_CARE_THRESHOLD_DEFAULT		= 0;	// x
	static const int nSTRAIGHT_THRESHOLD_DEFAULT		= 30;	// l
	static const int nPARTLY_BENT_THRESHOLD_DEFAULT		= 50;	// r
	static const int nCOMPLETELY_BENT_THRESHOLD_DEFAULT	= 75;	// n

public __gc class P5Glove : public IDisposable
	{
	private : 
		CP5Base* GloveCore;
		bool Disposed;
		bool initialized;

		// Helper variables added by Marc Eaddy - 5/7/05
		int* aryPositionThresholds;

		// Helper properties added by Marc Eaddy - 5/7/05
	public :

		__property int get_maxx()
		{
			return aryPositionThresholds[DontCare];
		}

		__property void set_maxx(int value)
		{
			aryPositionThresholds[DontCare] = value;
		}

		__property int get_maxl()
		{
			return aryPositionThresholds[Straight];
		}

		__property void set_maxl(int value)
		{
			aryPositionThresholds[Straight] = value;
		}

		__property int get_maxr()
		{
			return aryPositionThresholds[PartlyBent];
		}

		__property void set_maxr(int value)
		{
			aryPositionThresholds[PartlyBent] = value;
		}

		__property int get_maxn()
		{
			return aryPositionThresholds[TotallyBent];
		}

		__property void set_maxn(int value)
		{
			aryPositionThresholds[TotallyBent] = value;
		}

		bool IsGesture(	int thumbPos, 
						int indexPos, 
						int middlePos,
						int ringPos,
						int pinkyPos)
		{
			if (!initialized)
				return false;

			return	IsGesture(thumbPos) &&
					IsGesture(indexPos) &&
					IsGesture(middlePos) &&
					IsGesture(ringPos) && 
					IsGesture(pinkyPos);
		}

		bool IsGesture(int fingerState)
		{
			if (!initialized)
				return false;

			float fFingerValueToUse = 0;
			if (fingerState & Fingers::Thumb)
				fFingerValueToUse = GloveCore->Thumb;
			else if (fingerState & Fingers::Index)
				fFingerValueToUse = GloveCore->Index;
			else if (fingerState & Fingers::Middle)
				fFingerValueToUse = GloveCore->Middle;
			else if (fingerState & Fingers::Ring)
				fFingerValueToUse = GloveCore->Ring;
			else if (fingerState & Fingers::Pinky)
				fFingerValueToUse = GloveCore->Pinky;
			else
				return false; // Assert error!

			// Mask off the lower bits to get the index
			int positionIndex = fingerState & 0x0000FFFF;
			if (positionIndex < 0 || positionIndex > 4)
				return false; // Assert error!

			if (positionIndex == DontCare)
			{
				return true; // Don't care is always true
			}

			return fFingerValueToUse >= aryPositionThresholds[positionIndex-1] &&
				fFingerValueToUse < aryPositionThresholds[positionIndex];
		}

	public :

		__property float get_MouseX()
			{
			if (initialized)
				return GloveCore->mX;
			else
				return 0.0f;
			}

		__property float get_MouseY()
			{
			if (initialized)
				return GloveCore->mY;
			else
				return 0.0f;
			}

		__property float get_MouseZ()
			{
			if (initialized)
				return GloveCore->mZ;
			else
				return 0.0f;
			}

		__property float get_AbsoluteRoll()
		{
			if (initialized)
				return GloveCore->absRoll;
			else
				return 0.0f;
		}

		__property float get_RelativeRoll()
		{
			if (initialized)
				return GloveCore->relRoll;
			else
				return 0.0f;
		}

		__property float get_AbsoluteYaw()
		{
			if (initialized)
				return GloveCore->absYaw;
			else
				return 0.0f;
		}

		__property float get_RelativeYaw()
		{
			if (initialized)
				return GloveCore->relYaw;
			else
				return 0.0f;
		}

		__property float get_AbsolutePitch()
		{
			if (initialized)
				return GloveCore->absPitch;
			else
				return 0.0f;
		}

		__property float get_RelativePitch()
		{
			if (initialized)
				return GloveCore->relPitch;
			else
				return 0.0f;
		}

		__property int get_X()
		{
			if (initialized)
				return GloveCore->pX;
			else
				return 0;
		}

		__property int get_Y()
		{
			if (initialized)
				return GloveCore->pY;
			else
				return 0;
		}

		__property int get_Z()
		{
			if (initialized)
				return GloveCore->pZ;
			else
				return 0;
		}

		__property float get_Index()
		{
			if (initialized)
				return GloveCore->Index;
			else
				return 0.0f;
		}

		__property float get_Thumb()
		{
			if (initialized)
				return GloveCore->Thumb;
			else
				return 0.0f;
		}

		__property float get_Middle()
		{
			if (initialized)
				return GloveCore->Middle;
			else
				return 0.0f;
		}

		__property float get_Ring()
		{
			if (initialized)
				return GloveCore->Ring;
			else
				return 0.0f;
		}

		__property float get_Pinky()
		{
			if (initialized)
				return GloveCore->Pinky;
			else
				return 0.0f;
		}

		__property Fingers get_LeftButton()
		{
			Fingers ret = Fingers::Index;
			if (initialized)
			{
				GloveCore->MouseButtonAllocation();
				switch (GloveCore->LeftBtn)
				{
				case 0x00:
					ret = Fingers::Thumb;
					break;
				case 0x01:
					ret = Fingers::Index;
					break;
				case 0x02:
					ret = Fingers::Middle;
					break;
				case 0x03:
					ret = Fingers::Ring;
					break;
				case 0x04:
					ret = Fingers::Pinky;
					break;
				}
			}
			return ret;
		}

		__property Fingers get_MiddleButton()
		{
			Fingers ret = Fingers::Index;
			if (initialized)
			{
				GloveCore->MouseButtonAllocation();
				switch (GloveCore->MiddleBtn)
				{
				case 0x00:
					ret = Fingers::Thumb;
					break;
				case 0x01:
					ret = Fingers::Index;
					break;
				case 0x02:
					ret = Fingers::Middle;
					break;
				case 0x03:
					ret = Fingers::Ring;
					break;
				case 0x04:
					ret = Fingers::Pinky;
					break;
				}
			}
			return ret;
		}

		__property Fingers get_RightButton()
		{
			Fingers ret = Fingers::Index;
			if (initialized)
			{
				GloveCore->MouseButtonAllocation();
				switch (GloveCore->RightBtn)
				{
				case 0x00:
					ret = Fingers::Thumb;
					break;
				case 0x01:
					ret = Fingers::Index;
					break;
				case 0x02:
					ret = Fingers::Middle;
					break;
				case 0x03:
					ret = Fingers::Ring;
					break;
				case 0x04:
					ret = Fingers::Pinky;
					break;
				}
			}
			return ret;
		}

		__property bool get_Initialized()
		{
			return initialized;
		}

		__property bool get_MouseState()
		{
			if (initialized)
			return GloveCore->GetMouseState();
			else
				return false;
		}

		__property void set_MouseState(bool st)
		{
			if (initialized)
				GloveCore->SetMouseState(st);
		}

		__property bool get_ButtonA()
		{
			if (initialized)
				return GloveCore->GetButton(0);
			else
				return false;
		}

		__property bool get_ButtonB()
		{
			if (initialized)
				return GloveCore->GetButton(1);
			else
				return false;
		}

		__property bool get_ButtonC()
		{
			if (initialized)
				return GloveCore->GetButton(2);
			else
				return false;
		}

		__property bool get_ButtonD()
		{
			if (initialized)
				return GloveCore->GetButton(3);
			else
				return false;
		}

		__property bool get_InvertedX()
		{
			if (initialized)
				return GloveCore->IsInvertdX();
			else
				return false;
		}

		__property void set_InvertedX(bool inv)
		{
			if (initialized)
				GloveCore->InvertMouseX(inv);
		}

		__property bool get_InvertedY()
		{
			if (initialized)
				return GloveCore->IsInvertdY();
			else
				return false;
		}

		__property void set_InvertedY(bool inv)
		{
			if (initialized)
				GloveCore->InvertMouseY(inv);
		}

		__property bool get_InvertedZ()
		{
			if (initialized)
				return GloveCore->IsInvertdZ();
			else
				return false;
		}

		__property void set_InvertedZ(bool inv)
		{
			if (initialized)
				GloveCore->InvertMouseZ(inv);
		}

		P5Glove(void);
		P5Glove(int index);
		~P5Glove(void);
		void Close(void);
		void InitDevice(int index);
		void Process(void);
		void SetClipRegion(int xStart, int xEnd, int yStart, int yEnd, int zStart, int zEnd);
		void SetClipRegion(Control* control);
		void SetClickSensitivity(Fingers f, int val);
		void FastStartTracker(Control* control, int index);
		void FastStartTracker(Control* control);
		void Dispose(void);
		void CalibrateSensors(void);
		void CalibratePosition(void);
		void SaveSensors(void);

	};
}
