///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2005 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Engine/Devices/InputDeviceMapper.cs 8     1/12/07 2:23p Eaddy $
//
//    File name:	InputDeviceMapper.cs
//   Created by:	John Waugh
//   Created on:	04/08/2005
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Engine/Devices/InputDeviceMapper.cs $
 * 
 * 8     1/12/07 2:23p Eaddy
 * o After annotating all statements with their corresponding requirements
 * 
 * 7     10/10/06 3:15p Eaddy
 * o Removed redundant concern annotations
 * o We now have 100% concern annotation coverage
 * 
 * 6     9/29/06 6:18p Eaddy
 * o Annotated program elements with the concern they realize
 * 
 * 5     5/02/06 11:03a Eaddy
 * o Moved MyTrace into separate library
 * 
 * 4     5/28/05 1:20p Eaddy
 * Changed copyright
 * 
 * 3     5/07/05 12:48p Eaddy
 * o Added header comment
 * o Added Dispose() so that InterSense socket can be released
 */

using Phx.Aop.Attributes;

using System;
using System.Collections;
using System.Xml;
using System.Xml.Serialization;
using System.Runtime.Serialization;
using System.Diagnostics;

using Columbia.Utilities;

namespace Goblin.Engine.Devices
{
	/// <summary>
	/// This class performs two functions
	///		1) it provides a mapping from string identifier to InputDevice, via its indexer
	///		2) it contains a list of all InputDeviceDrivers, and serializes them when it itself is serialized
	/// </summary>
	/// 

	[Feature("Input")]
	[Requirement("Configuration.Device configuration files (#135)")]
	[Requirement("Device.Device abstraction (#100)")]
	[Requirement("Device.Dynamic device discovery (#105)")]
	public class InputDeviceMapper : ICustomSerialization, IInitAfterDeserialization, IDisposable
	{
		/// <summary>
		/// The InputDeviceMapper begins with no knowledge of drivers, you must add them either via Deserialize
		/// </summary>
		public InputDeviceMapper()
		{
		}

		[Requirement("Nonfunctional.Clean shutdown (#1405)")]
		public void Dispose()
		{
			foreach(InputDeviceDriver idd in drivers)
			{
				idd.Dispose();
			}
		}

		[Feature("Persistence")]
		[Requirement("Configuration.Device configuration files (#135)")]
		public bool HasChanged
		{
			get { return hasChanged; }
		}

		/// <summary>
		/// Implementation of ICustomSerialization
		/// NOTE: do not try to use XmlSerializer.Serialize directly on this object
		///		this object has to run special code to deal with the homogeneous types in its list of InputDeviceDrivers
		/// </summary>
		/// <param name="writer"></param>
		[Feature("Persistence.Saving")]
		[Requirement("Configuration.Device configuration files (#135)")]
		[Requirement("Configuration.XML (#400)")]
		public void Serialize(XmlWriter writer, Type[] extraTypes)
		{
			/*
			 * since we don't know what subclasses of InputDeviceDriver may be in our devices ArrayList,
			 * we must manually traverse the list and add the corresponding types to the list of overrides
			 * since XmlSerializer needs to know the types of the objects it is serializing
			 */

			XmlAttributes xAttrs = new XmlAttributes();

			for(int i=0;i<drivers.Count;i++)
			{
				XmlArrayItemAttribute xArrayItem = 
					new XmlArrayItemAttribute(drivers[i].GetType());
				xAttrs.XmlArrayItems.Add(xArrayItem);
			}

			// Add all overrides to XmlAttribueOverrides object
			XmlAttributeOverrides xOver = new XmlAttributeOverrides();
			xOver.Add(typeof(InputDeviceMapper), "drivers", xAttrs);

			// Create the XmlSerializer, with the overrides
			XmlSerializer s = new XmlSerializer(typeof(InputDeviceMapper), 
				xOver, extraTypes, null, null);

			s.Serialize(writer,this);
		}

		/// <summary>
		/// This function is called by XmlSerializerHelper, if you choose to use it
		/// 
		/// This returns a deserialized InputDeviceMapper.
		/// driverTypes is an array of all the types of driver you
		/// are interested in having 're-hydrated' from the XML file
		/// </summary>
		/// <param name="driverTypes"></param>
		/// <returns></returns>
		[Feature("Persistence.Loading")]
		[Requirement("Configuration.Device configuration files (#135)")]
		[Requirement("Device.Device abstraction (#100)")]
		[Requirement("Device.Dynamic device discovery (#105)")]
		[Requirement("Rendering.Coordinate system independence (#715)")]
		public static InputDeviceMapper Deserialize(XmlReader reader, Type[] driverTypes)
		{
			XmlAttributes xAttrs = new XmlAttributes();

			for(int i=0; i < driverTypes.Length; i++)
			{
				XmlArrayItemAttribute xArrayItem = new XmlArrayItemAttribute(driverTypes[i]);
				xAttrs.XmlArrayItems.Add(xArrayItem);
			}

			// Add all overrides to XmlAttribueOverrides object
			XmlAttributeOverrides xOver = new XmlAttributeOverrides();
			xOver.Add(typeof(InputDeviceMapper), "drivers", xAttrs);

			// Create the XmlSerializer, with the overrides
			XmlSerializer s = new XmlSerializer(typeof(InputDeviceMapper), xOver);
			
			//InitAfterDeserialization will be called by XmlSerializerHelper (or whoever the outside entity is)
			return (InputDeviceMapper) s.Deserialize(reader);
		}
		
		/// <summary>
		/// Implementation of IInputAfterDeserialization
		/// </summary>
		[Feature("Persistence.Loading")]
		public void InitAfterDeserialization()
		{
			//we have just finished being de-serialized
			//do custom initialzation (build the hash)

			foreach(InputDeviceDriver idd in drivers)
			{
				driverHash.Add(idd.identifier,idd);
				idd.InitAfterDeserialization();
			}

			//start monitoring config changes from here
			hasChanged = false;
		}

		/// <summary>
		/// main accessor for this class
		/// takes a unique device identifier string and 
		/// returns the corresponding InputDeviceDriver (read only)
		/// </summary>
		/// 
		public InputDeviceDriver this[string device_id]
		{
			get
			{
				if (driverHash.Contains(device_id))
					return driverHash[device_id] as InputDeviceDriver;
				else
					return null;
			}
		}

		/// <summary>
		/// Call this from your main loop.  This will update all the drivers that are mapped.
		/// </summary>
		/// <param name="elapsed_seconds"></param>
		public void Update(float elapsed_seconds)
		{
			foreach(InputDeviceDriver d in drivers)
				d.Update(elapsed_seconds);
		}

		/// <summary>
		/// convenience method - prints out a list of the names of all drivers
		/// followed by a list of all inputs
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			IDictionaryEnumerator e = driverHash.GetEnumerator();
			string ret = "Drivers: ";
			if(e.MoveNext())
			{
				ret += "["+e.Key.ToString();
				while(e.MoveNext())
				{
					ret += ", " + e.Key;
				}

				ret += "]";
			}
			else
				ret += "<empty>";
			
			return ret;
		}

		//this stores all our InputDeviceDriver instances
		//DO NOT ACCESS THIS EXTERNALLY, even though it's public (has to be for serialization purposes)
		[XmlArray]
		[XmlArrayItem(typeof(Object))]
		[Feature("Persistence")]
		public ArrayList drivers = new ArrayList(32);

		private Hashtable driverHash = new Hashtable(32);

		//indicates if the configuration has changed since object creation
		[Feature("Persistence")]
		private bool	hasChanged = false;
	}
}
