///////////////////////////////////////////////////////////////////////
//  Copyright (C) 2005 Columbia University.  All Rights Reserved.
//
//      $Header: /Goblin/Engine/Scene/Bounds.cs 8     1/12/07 2:23p Eaddy $
//
//    File name:	Bounds.cs
//   Created by:	Marc Eaddy
//   Created on:	05/16/2005
//
//      Purpose:	
//
//  Description:	
//
///////////////////////////////////////////////////////////////////////
/**
 * $Log: /Goblin/Engine/Scene/Bounds.cs $
 * 
 * 8     1/12/07 2:23p Eaddy
 * o After annotating all statements with their corresponding requirements
 * 
 * 7     9/29/06 6:18p Eaddy
 * o Annotated program elements with the concern they realize
 * 
 * 6     5/02/06 11:03a Eaddy
 * o Moved MyTrace into separate library
 * 
 * 5     5/31/05 5:49p Eaddy
 * o Added Scale()
 * o 
 * 
 * 4     5/31/05 2:56p Eaddy
 * o Can now render itself (this is expensive so use it sparingly)
 * o Now inflates the vol slightly to compensate for the too small vol
 * generated
 * by Geometry.ComputeBoundingXXX
 * 
 * 3     5/30/05 11:51p Eaddy
 * o Fixed bug that was introduced when I added the walls
 * 
 * 2     5/28/05 1:21p Eaddy
 * Changed copyright
 * 
 * 1     5/17/05 10:46p Eaddy
 * o Moved Bounds class into separate file
 */

using Phx.Aop.Attributes;

using System;

using System.Diagnostics;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

using Columbia.Utilities;

using Goblin.Engine.Utilities;

namespace Goblin.Engine.Scene
{
	[Feature("SceneGraph.Geometry")]
	[Requirement("App Development.Collision detection (#1150)")]
	[Requirement("Debugging.Application visualization (#620)")]
	public class Bounds : IDisposable
	{
		// DATA

		// Bounding sphere variables

		/// <summary>
		/// Center of bounding sphere of object
		/// </summary>
		public Vector3 center;

		/// <summary>
		/// Radius of bounding sphere of object
		/// </summary>
		public float radius;

		//Axis-aligned bounding box (AABB) variables

		public Vector3 min;
		public Vector3 max;

		public bool boundingSphere;

		/// <summary>
		/// Random number generator for obtaining
		/// waypoints and spawn points
		/// </summary>
		Random rand = null;

		/// <summary>
		/// Mesh for the bounding volume
		/// </summary>
		Mesh mesh = null;

		readonly float VOLUME_FACTOR = (float)Math.PI * 4.0f/3.0f;
		readonly float SCALE_FACTOR = .2f; // Percentage of original bounding volume

		public Bounds()
		{
			center = Vector3.Empty;
			radius = 0.0f;
			min = Vector3.Empty;
			max = Vector3.Empty;
			boundingSphere = false;
			mesh = null;
		}

		//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
		public void Dispose()
		{
			DisposeMesh();
		}

		//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
		public void DisposeMesh()
		{
			if (mesh != null)
			{
				mesh.Dispose();
				mesh = null;
			}
		}

		public Bounds set(Bounds that)
		{
			//[Requirement("Debugging.Exception/error detection (#625)")]
			if (this == that)
				//[Requirement("Debugging.Exception/error handling (#630)")]
				return this;

			center = that.center;
			radius = that.radius;
			min = that.min;
			max = that.max;
			boundingSphere = that.boundingSphere;
			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			DisposeMesh();

			return this;
		}

		public void setAABB(Vector3 min, Vector3 max)
		{
			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			DisposeMesh(); // Mesh no longer valid

			boundingSphere = false;
			this.min = min;
			this.max = max;
		}

		public void setSphere(Vector3 center, float radius)
		{
			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			DisposeMesh(); // Mesh no longer valid

			boundingSphere = true;
			this.center = center;
			this.radius = radius;
		}

		private void add(Bounds bounds)
		{
			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			DisposeMesh(); // Mesh no longer valid

			min.X = Math.Min(min.X, bounds.min.X);
			min.Y = Math.Min(min.Y, bounds.min.Y);
			min.Z = Math.Min(min.Z, bounds.min.Z);

			max.X = Math.Max(max.X, bounds.max.X);
			max.Y = Math.Max(max.Y, bounds.max.Y);
			max.Z = Math.Max(max.Z, bounds.max.Z);

			// This does not work!
			Vector3 middle = center + bounds.center;
			middle.Scale(0.5f);
			float dist = (center - middle).Length();
			radius += dist;
		}

		public void update(GraphicsStream vertStream, int numVerts, VertexFormats format)
		{
			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			DisposeMesh(); // Mesh no longer valid

			Geometry.ComputeBoundingBox(vertStream, numVerts, format, 
				out min, out max);
			
			radius = Geometry.ComputeBoundingSphere(vertStream, numVerts, 
				format, out center);

			SPHEREvsBOX();
		}

		public void update(GraphicsStream vertStream, int numVerts)
		{
			update(vertStream, numVerts, VertexFormats.Position);
		}

		public void update(Vector3[] vertices)
		{
			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			DisposeMesh(); // Mesh no longer valid

			Geometry.ComputeBoundingBox(vertices, VertexFormats.Position, 
				out min, out max);

			radius = Geometry.ComputeBoundingSphere(vertices, 
				VertexFormats.Position, out center);

			SPHEREvsBOX();
		}

		private void SPHEREvsBOX()
		{
			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			DisposeMesh(); // Mesh no longer valid

			float sphereVolume = VOLUME_FACTOR * radius * radius * radius;
			Vector3 tmp = max - min;
			float boxVolume = tmp.X * tmp.Y * tmp.Z;

            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(min.X <= max.X && min.Y <= max.Y && min.Z <= max.Z);

			boundingSphere = sphereVolume < boxVolume;
		}

		private void Scale(float factor)
		{
			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			DisposeMesh(); // Mesh no longer valid

			float sphereVolume = VOLUME_FACTOR * radius * radius * radius;
			Vector3 tmp = max - min;
			float boxVolume = tmp.X * tmp.Y * tmp.Z;

            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(min.X <= max.X && min.Y <= max.Y && min.Z <= max.Z);

			// Scale sphere
			sphereVolume *= SCALE_FACTOR + 1f;
			radius = (float) Math.Pow(sphereVolume / VOLUME_FACTOR, 1.0/3.0);

			// Scale box
			tmp *= SCALE_FACTOR / 2f;
			min -= tmp;
			max += tmp;
		}

		public Vector3 getRandomPosition()
		{
			if (rand == null)
				rand = new Random();

			// We don't handle sphere's yet
			//[Requirement("Debugging.Exception/error detection (#625)")]
			if (boundingSphere)
				//[Requirement("Debugging.Exception/error handling (#630)")]
				return Vector3.Empty;

			return new Vector3(
				((float) rand.NextDouble() * (max.X - min.X)) + min.X,
				((float) rand.NextDouble() * (max.Y - min.Y)) + min.Y,
				((float) rand.NextDouble() * (max.Z - min.Z)) + min.Z);
		}

		public Mesh getMesh(Device device)
		{
			if (mesh != null)
				return mesh;

			Matrix translate;

			if (boundingSphere)
			{
				//[Requirement("Debugging.Exception/error detection (#625)")]
				if (radius == 0)
					//[Requirement("Debugging.Exception/error handling (#630)")]
					return null;

				mesh = Mesh.Sphere(device, radius, 10, 10);
				translate = Matrix.Translation(center);
			}
			else
			{
				Vector3 dimensions = max - min;

				//[Requirement("Debugging.Exception/error detection (#625)")]
				if (dimensions.X == 0 &&
					dimensions.Y == 0 &&
					dimensions.Z == 0)
					//[Requirement("Debugging.Exception/error handling (#630)")]
					return null;

				// Create a box with the right dimensions
				mesh = Mesh.Box(device, dimensions.X, dimensions.Y, dimensions.Z);

				// Calculate the translation

				// The box is originally centered at the origin so move it until
				// the corner is at the origin. The origin in this case is the
				// min point.
				dimensions.Scale(0.5f); // Move each dimension by 1/2
				translate = Matrix.Translation(min + dimensions);
			}

			// Transform the vertices
			VertexBuffer vbTrans = mesh.VertexBuffer;

            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(vbTrans.Description.VertexFormat == CustomVertex.PositionNormal.Format);

			GraphicsStream strmDest = vbTrans.Lock(0, 0, LockFlags.None);

			unsafe
			{
				CustomVertex.PositionNormal* pVertsDest = 
					(CustomVertex.PositionNormal*) strmDest.InternalDataPointer;

				for(int i = 0; i < mesh.NumberVertices; ++i)
				{
					pVertsDest[i].Position = 
						Vector3.TransformCoordinate(pVertsDest[i].Position, translate);
					//pVertsDest[i].Normal = 
					//	Vector3.TransformCoordinate(pVertsDest[i].Normal, translate);
				}
			}

			//[Requirement("Nonfunctional.Clean shutdown (#1405)")]
			{
				strmDest.Dispose();
	
				vbTrans.Unlock();
				vbTrans.Dispose();
			}

			return mesh;
		}

		public void Render(Device device)
		{
			Mesh mesh = getMesh(device);
			//[Requirement("Debugging.Exception/error detection (#625)")]
			if (mesh != null)
				mesh.DrawSubset(0);
		}

		public bool Contains(Vector3 pt)
		{
			if (boundingSphere)
			{
				Vector3 diff = pt - center;
				return diff.Length() <= radius;
			}
			else
			{
				return	pt.X >= min.X && pt.X <= max.X &&
						pt.Y >= min.Y && pt.Y <= max.Y &&
						pt.Z >= min.Z && pt.Z <= max.Z;
			}
		}

		[Requirement("Nonfunctional.Unused (#1400)")]
		public override string ToString()
		{
			return string.Format("{0}: Rad={1} Center={2} Min={3} Max={4}",
				boundingSphere ? "Sphere" : "Box",
				radius,
				MyUtility.Vector3ToString(center),
				MyUtility.Vector3ToString(min),
				MyUtility.Vector3ToString(max));
		}

		[Requirement("App Development.Collision detection (#1150)")]
		public static bool Overlaps(Bounds bounds1, Bounds bounds2)
		{
            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(bounds1 != null && bounds2 != null);

			if ( bounds1.boundingSphere )
			{
				if ( bounds2.boundingSphere )
				{
					return SpheresOverlap(bounds1, bounds2);
				}
				else
				{
					return BoxAndSphereOverlaps(bounds2, bounds1);
				}
			}
			else if ( bounds2.boundingSphere )
			{
				return BoxAndSphereOverlaps(bounds1, bounds2);
			}
			else
			{
				return BoxesOverlap(bounds1, bounds2);
			}
		}

		[Requirement("App Development.Collision detection (#1150)")]
		private static bool SpheresOverlap(Bounds bounds1, Bounds bounds2)
		{
            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(bounds1.boundingSphere);
            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(bounds2.boundingSphere);

			float radiusSqr = bounds1.radius + bounds2.radius;
			radiusSqr *= radiusSqr;
			Vector3 dist = bounds2.center - bounds1.center;
			return (dist.LengthSq() - radiusSqr) < 0;
		}

		[Requirement("App Development.Collision detection (#1150)")]
		private static bool BoxesOverlap(Bounds bounds1, Bounds bounds2)
		{
            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(!bounds1.boundingSphere);
            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(!bounds2.boundingSphere);

			if (bounds1.min.X > bounds2.max.X ||
				bounds1.min.Y > bounds2.max.Y ||
				bounds1.min.Z > bounds2.max.Z ||
				bounds2.min.X > bounds1.max.X ||
				bounds2.min.Y > bounds1.max.Y ||
				bounds2.min.Z > bounds1.max.Z )
				return false;
			else 
				return true;
		}

		[Requirement("App Development.Collision detection (#1150)")]
		private static bool BoxAndSphereOverlaps(Bounds bounds1, Bounds bounds2)
		{
            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(!bounds1.boundingSphere);
            //[Requirement("Debugging.Exception/error detection (#625)")]
            //[Requirement("Debugging.Exception/error handling (#630)")]
			Debug.Assert(bounds2.boundingSphere);

			if (IsSphereInBox(bounds1.min, bounds1.max, bounds2.center)) 
				return true;

			Vector3 max = bounds1.min - bounds2.center;

			if (max.X > bounds2.radius ||
				max.Y > bounds2.radius ||
				max.Z > bounds2.radius)
				return false;

			Vector3 min = bounds2.center - bounds1.max;

			if (min.X > bounds2.radius ||
				min.Y > bounds2.radius || 
				min.Z > bounds2.radius ) 
				return false;
			else
				return true;
		}

		[Requirement("App Development.Collision detection (#1150)")]
		private static bool IsSphereInBox(Vector3 min, Vector3 max, Vector3 center )
		{
			if (min.X > center.X ||
				min.Y > center.Y ||
				min.Z > center.Z ||
				center.X > max.X ||
				center.Y > max.Y ||
				center.Z > max.Z ) 
				return false;
			else
				return true;
		}
	}
}
